#include "GrilleSudoku.h"
#include <stdlib.h>
#include <time.h>
#include <QRegExpValidator>
#include <QProgressDialog>

#include <QtGui>

GrilleSudoku::GrilleSudoku()
{
	int i,j;
	niveau = 2;
	jeu_termine = false;
	jeu_resolu = false;
	delaiOK = true;

	//initialisation des couleurs
	genererColor = new QColor(255, 62, 36);
	aideColor = new QColor(41, 212, 255);
	resoluColor = new QColor(42, 255, 92);
	normalColor = new QColor(255, 255, 255);

	delai = new QTimer(this);
	delai->setSingleShot(true);
	connect(delai, SIGNAL(timeout()), this, SLOT(delaiPasse()));

	for(i=0;i<9;i++)
    {
		for(j=0;j<9;j++)
    	{
        	m[i][j]=0;
        	status[i][j]=0;
        	mLine[i][j] = new QLineEdit();
        	mLine[i][j]->setObjectName(tr("case%1").arg(i*10+j));
        	mLine[i][j]->installEventFilter(this);
        	mLine[i][j]->setAlignment(Qt::AlignCenter);
        	mLine[i][j]->setMaximumSize(30,30);
        	mLine[i][j]->setMinimumSize(30,30);
        	QRegExp rx("[1-9]");
			mLine[i][j]->setValidator(new QRegExpValidator(rx, this)); //permet de ne pouvoir saisir que des nombre de 1  9

    		connect(mLine[i][j], SIGNAL(editingFinished()),  this, SLOT(setNbPossibles()));
		}
    }


	grilleLayout = new QGridLayout();

	for(i=0;i<9;i++)
	{
		for(j=0;j<9;j++)
		{
			grilleLayout->addWidget(mLine[i][j],i+i/3,j+j/3);
		}
	}

	grilleLayout->setColumnMinimumWidth( 3, 15);
	grilleLayout->setColumnMinimumWidth( 7, 15);
	grilleLayout->setRowMinimumHeight( 3, 15);
	grilleLayout->setRowMinimumHeight( 7, 15);

	grilleLayout->setSizeConstraint(QLayout::SetFixedSize);

	setLayout(grilleLayout);

}
GrilleSudoku::~GrilleSudoku()
{

}

bool GrilleSudoku::generer()
{
	int i,j,k,n;
	int nb_case=0;
	bool cancel=false;

	delaiOK = true;

	srand(time(NULL));

	for(i=0;i<9;i++)for(j=0;j<9;j++) m_save[i][j]=m[i][j]; //sauvegarde matrice

	genererSudoku(); //cre une grille alatoire

	switch (niveau)
	{
		case 1 : nb_case=60; break;
		case 2 : nb_case=90; break;
		case 3 : nb_case=70; break;
	}

	QProgressDialog progress("Gnration d'une nouvelle grille...", "Stop", 0, 160, this);
	progress.setValue(0);
	progress.setWindowTitle("SudoKU");
	progress.show();
	for(k=0;k<nb_case && !cancel;k++)
	{
		i=rand()%9;
		j=rand()%9;
		if (m[i][j]!=0)
		{
			progress.setValue(k);
			qApp->processEvents();
			if (progress.wasCanceled()) cancel=true;
			else
			{
				n=m[i][j];
				m[i][j]=0;
				if (nbSolutions(nbCasesVides(),0)>1) m[i][j]=n;
				//if (nbSolutions3(nbCasesVides(),0)>1) m[i][j]=n;
				//if (nbSolutions2(nbCasesVides(),0,0,0)>1) m[i][j]=n;
			}
		}
	}
	if (niveau==3)
	{
		for (i=0;i<9 && !cancel;i++)
		{
			for (j=0;j<9 && !cancel;j++)
			{
				if (m[i][j]!=0)
				{
					progress.setValue(79+i*9+j);
					qApp->processEvents();
					if (progress.wasCanceled()) cancel=true;
					else
					{
						n=m[i][j];
						m[i][j]=0;
						if (nbSolutions(nbCasesVides(),0)>1) m[i][j]=n;
						//if (nbSolutions3(nbCasesVides(),0)>1) m[i][j]=n;
						//if (nbSolutions2(nbCasesVides(),0,0,0)>1) m[i][j]=n;
					}
				}
			}
		}
	}

	if (!cancel)
	{
		for(i=0;i<9;i++)
		{
			for(j=0;j<9;j++)
			{
				if (m[i][j] == 0) status[i][j]=0;
				else status[i][j]=1;

				m_debut[i][j] = m[i][j];
				setMLine(m[i][j],i,j);
			}
		}

		jeu_termine = false;
		jeu_resolu = false;

		return true;
	}
	else
	{
		for(i=0;i<9;i++) for(j=0;j<9;j++) m[i][j]=m_save[i][j]; //restauration matrice
		return false;
	}
}

bool GrilleSudoku::resoudre()
{
	int i,j;

	if (!testNbSolutions()) return false;


	jeu_resolu=false;
    resoudreRecursivement(nbCasesVides(),0);

    if(jeu_resolu)
    {

		for(i=0;i<9;i++)
		{
			for(j=0;j<9;j++)
			{
				if (m[i][j]==0)
				{
					status[i][j]=2;
					setMLine(m_resultat[i][j],i,j);
				}
			}
		}
		return true;
	}
	else
	{
		QMessageBox::warning(this, tr("Erreur"),tr("Cette grille est incorrecte, elle ne comporte pas de solution"));
		return false;
	}
}

void GrilleSudoku::aide()
{
    int i, j, nb=0,r;
    int liste_i[65];
    int liste_j[65];

    srand(time(NULL));

    if (!testNbSolutions()) return;

    for (i=0;i<9;i++)
    {
		for (j=0;j<9;j++)
    	{
			if (mLine[i][j]->text()=="")
			{
				liste_i[nb]=i;
				liste_j[nb]=j;
				nb++;
			}
		}
	}

    if (nb>0)
    {
		jeu_resolu=false;
		resoudreRecursivement(nbCasesVides(),0);

		if(jeu_resolu)
		{
			r = rand()%nb;
			i = liste_i[r];
			j = liste_j[r];
			status[i][j]=3;
			setMLine(m_resultat[i][j],i,j);
			m[i][j] = m_resultat[i][j];
		}
		else
		{
			QMessageBox::warning(this, tr("Erreur"),tr("Cette grille est incorrecte, elle ne comporte pas de solution"));
		}
		jeu_resolu=false;
	}
}

void GrilleSudoku::recommencer()
{
	int i,j;
	for(i=0;i<9;i++)
	{
		for(j=0;j<9;j++)
		{
			m[i][j]=m_debut[i][j];
			if(m[i][j]==0) status[i][j]=0;
			else status[i][j]=1;
			setMLine(m[i][j],i,j);
		}
	}
	jeu_termine = false;
	jeu_resolu = false;
}

//void GrilleSudoku::print();

void GrilleSudoku::load(const QString &nomDuFichier, bool nouveau)
{
	int i,j,n;
	QFile file(nomDuFichier);
	if (!file.open(QFile::ReadOnly | QFile::Text)) {
		QMessageBox::warning(this, tr("Erreur"),tr("Impossible de lire le fichier %1:\n%2.").arg(nomDuFichier).arg(file.errorString()));
		return;
	}

	QTextStream in(&file);
	QString *str = new QString(in.readAll());

	for(i=0;i<9;i++)
	{
		for(j=0;j<9;j++)
		{
			status[i][j] = str->mid((i*9+j)*2,1).toInt();
			n=str->mid((i*9+j)*2+1,1).toInt();
			if(status[i][j]==1)
			{
				m[i][j]=n;
				if(nouveau) status[i][j]=0;
			}
			else m[i][j]=0;
			m_debut[i][j] = m[i][j];
			setMLine(n,i,j);
		}
	}
}
bool GrilleSudoku::save(const QString &nomDuFichier, bool nouveau)
{
	int i,j;
	QFile file(nomDuFichier);
	if (!file.open(QFile::WriteOnly | QFile::Text)) {
		QMessageBox::warning(this, tr("Erreur"),tr("Impossible d'crire le fichier %1:\n%2.").arg(nomDuFichier).arg(file.errorString()));
		return false;
	}
	QTextStream out(&file);

	for(i=0;i<9;i++)
	{
		for(j=0;j<9;j++)
		{
			if(nouveau)
			{
				if (mLine[i][j]->text()=="") out << "00";
				else out << "1" + mLine[i][j]->text();
			}
			else
			{
				out << (char) (status[i][j]+'0');
				if (mLine[i][j]->text()=="") out << "0";
				else out << mLine[i][j]->text();
			}
		}
	}
	return true;
}
bool GrilleSudoku::saveHtml(const QString &nomDuFichier)
{
	int i,j;

	QFile file(nomDuFichier);
	if (!file.open(QFile::WriteOnly | QFile::Text)) {
		QMessageBox::warning(this, tr("Erreur"),tr("Impossible d'crire le fichier %1:\n%2.").arg(nomDuFichier).arg(file.errorString()));
		return false;
	}

	QTextStream out(&file);

	out << "<html><head><title>SudoKU</title></head><body><table border=\"1\">";

	for (i=0;i<9;i++)
	{
		out << "<tr>";
		for (j=0;j<9;j++)
		{
			out << "<td>&nbsp;";
			if (mLine[i][j]->text()=="")
				out << "&nbsp;";
			else
				out << mLine[i][j]->text();
			out << "&nbsp;</td>";
		}
		out << "</tr>";

	}
	out << "</table></body></html>";
	return true;
}

void GrilleSudoku::nouveau()
{
	int i,j;
	for (i=0;i<9;i++)
	{
		for(j=0;j<9;j++)
		{
			m[i][j]=status[i][j]=0;
			niveau = 2;
			jeu_termine = false;
			jeu_resolu = false;
			delaiOK = true;

			setMLine(0,i,j);
		}
	}
}

bool GrilleSudoku::tester()
{
	bool resultat = true;
	int i, j, k;
	bool liste_nb[9];
	int liste[3][9];

	for (i=0;i<9;i++) for(j=0;j<9;j++) {m_save[i][j] = m[i][j]; m[i][j]=mLine[i][j]->text().toInt();}

	for (i=0;i<9;i++)
	{
		ligne(liste[0],i);
		colonne(liste[1],i);
		zone(liste[2],(i/3)*3,(i%3)*3);
		for(j=0;j<3;j++)
		{
			for(k=0;k<9;k++) liste_nb[k]=true;
			for(k=0;k<9;k++)
			{
				if(liste[j][k]!=0)
				{
					if(!liste_nb[liste[j][k]-1]) resultat=false;
					else liste_nb[liste[j][k]-1]=false;
				}
			}
		}
	}

	for (i=0;i<9;i++) for(j=0;j<9;j++) m[i][j] = m_save[i][j];

	return resultat;
}

bool GrilleSudoku::estTermine()
{
	int i,j;
	bool fini=true;

	for(i=0;i<9 && fini;i++) for(j=0;j<9 && fini;j++) if (mLine[i][j]->text()=="") fini=false;

	if (fini && !jeu_termine && !jeu_resolu)
	{
		resoudreRecursivement(nbCasesVides(),0);
		for(i=0;i<9 && fini;i++)
			for(j=0;j<9 && fini;j++)
				if (mLine[i][j]->text().toInt()!=m_resultat[i][j]) fini=false;
		if (fini)
		{
			jeu_termine=true;
			QMessageBox::information(this, "Bien jou !","Bravo, vous avez termin cette grille !");
		}
	}
	return fini;
}


int GrilleSudoku::getNiveau()
{
	return niveau;
}
void GrilleSudoku::setNiveau(int n)
{
	if(n>0 && n<4)
	{
		niveau = n;
	}
}

void GrilleSudoku::setNombresPossibles(bool affiche)
{
	int i,j,k,n;
	int liste[9];

	QString *s = new QString();

	nbPossibles = affiche;

	for (i=0;i<9;i++)
	{
		for (j=0;j<9;j++)
		{
			m_save[i][j] = m[i][j];
			m[i][j] = mLine[i][j]->text().toInt();
		}
	}

	if(affiche)
	{
		for (i=0;i<9;i++)
		{
			for (j=0;j<9;j++)
			{
				if(nombresPossibles(liste,i,j,&n))
				{
					s->clear();
					s->append("Nombres possibles :");
					for (k=0;k<n;k++)
					{
						s->append(tr(" %1").arg(liste[k]));
					}
					mLine[i][j]->setStatusTip(*s);
				}
				else
				{
					mLine[i][j]->setStatusTip("");
				}
			}
		}
	}
	else
	{
		for (i=0;i<9;i++)
		{
			for (j=0;j<9;j++)
			{
				mLine[i][j]->setStatusTip("");
			}
		}
	}

	for (i=0;i<9;i++)
	{
		for (j=0;j<9;j++)
		{
			m[i][j] = m_save[i][j];
		}
	}
}
bool GrilleSudoku::getNombresPossibles(){ return nbPossibles;}


bool GrilleSudoku::eventFilter(QObject *obj, QEvent *event)
{
	if(!obj->objectName().isNull() && !obj->objectName().isEmpty() )
	{
		if (obj->objectName().left(4) == "case")
		{
			if (event->type() == QEvent::KeyPress) //si on a appuy sur une touche
			{
				QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);

				int numCase = obj->objectName().mid(4).toInt();
				if(keyEvent->key()==Qt::Key_Up){if(numCase/10 == 0) numCase += 80;else numCase-=10;}
				if(keyEvent->key()==Qt::Key_Down){if(numCase/10 == 8) numCase -= 80;else numCase+=10;}
				if(keyEvent->key()==Qt::Key_Left){if(numCase%10 == 0) numCase += 8;else numCase --;}
				if(keyEvent->key()==Qt::Key_Right){if(numCase%10 == 8) numCase -= 8;else numCase ++;}

				findChild<QLineEdit *>(tr("case%1").arg(numCase))->setFocus();
				findChild<QLineEdit *>(tr("case%1").arg(numCase))->selectAll();
			}
		}
	}
	return QWidget::eventFilter(obj, event);
}


void GrilleSudoku::setColors(QColor g, QColor a, QColor r, QColor n)
{
	int i, j;

	genererColor->setRgb(g.rgb());
	aideColor->setRgb(a.rgb());
	resoluColor->setRgb(r.rgb());
	normalColor->setRgb(n.rgb());

	QPalette newPalette = palette();

	for (i=0;i<9;i++)
	{
		for (j=0;j<9;j++)
		{
			switch (status[i][j])
			{
			case 0 :
				newPalette.setColor(QPalette::Base, *normalColor);
				mLine[i][j]->setPalette(newPalette);
				break;
			case 1 :
				newPalette.setColor(QPalette::Base, *genererColor);
				mLine[i][j]->setPalette(newPalette);
				break;
			case 2 :
				newPalette.setColor(QPalette::Base, *resoluColor);
				mLine[i][j]->setPalette(newPalette);
				break;
			case 3 :
				newPalette.setColor(QPalette::Base, *aideColor);
				mLine[i][j]->setPalette(newPalette);
				break;
			}
		}
	}
}
void GrilleSudoku::getColors(QColor *g, QColor *a, QColor *r, QColor *n)
{
	g=genererColor;
	a=aideColor;
	r=resoluColor;
	n=normalColor;
}


/*************************************/
/***************private***************/
/*************************************/

void GrilleSudoku::genererSudoku()
{
	int liste[9];
	int nb;
	int i,j;
	bool echec=true;

	srand(time(NULL));
	while (echec)
	{
		for(i=0;i<9;i++) for(j=0;j<9;j++) m[i][j]=0;

		echec=false;
		for(i=0;i<9 && !echec;i++)
		{
			for(j=0;j<9 && !echec;j++)
			{
				nombresPossibles(liste,i,j,&nb);
				if(nb!=0)
				{
					//tirage alatoire d'un nombre possible
					m[i][j]=liste[rand()%nb];
				}
				else echec=true; //si tous les nombres sont utiliss : chec
			}
		}
	}
}


/*
int nbSolutions(int nb_a_resoudre, int nb_resolu)
Retourne le nombre de solutions de la grille
	int nb_a_resoudre	: nombre de cases  rsoudre. 	Au dbut : nombre de cases vide -> nbCasesVides()
	int nb_resolu		: nombre de cases rsolu.		Au dbut : 0
*/
int GrilleSudoku::nbSolutions(int nb_a_resoudre, int nb_resolu)
{
	int i=0,j=0,k,nb=0,n;
	int liste[9];
	bool solution=true;


	//qApp->processEvents();

	if (!delaiOK) return 0;

	if (nb_a_resoudre>nb_resolu)
	{
		for (i=0;i<9 && solution;i++)
		{
			for (j=0;j<9 && solution;j++)
			{
				if (nombresPossibles(liste, i, j, &n)) //si m[i][j]==0
				{
					solution=false;
					for (k=0;k<n;k++)
					{
						m[i][j]=liste[k];
						nb += nbSolutions(nb_a_resoudre,nb_resolu+1);
						m[i][j]=0;
					}
				}
			}
		}
	}
	else nb++;
	return nb;
}
int GrilleSudoku::nbSolutions3(int nb_a_resoudre, int nb_resolu)
{
	int i=0,j=0,k,nb=0,n;
	int liste[9];

	qApp->processEvents();

	if (!delaiOK) return 0;
	if(nb_a_resoudre==nb_resolu) return 1;


	for (i=0;i<9;i++)
	{
		for (j=0;j<9;j++)
		{
			if (nombresPossibles(liste, i, j, &n)) //si m[i][j]==0
			{
				for (k=0;k<n;k++)
				{
					m[i][j]=liste[k];
					nb += nbSolutions3(nb_a_resoudre,nb_resolu+1);
					m[i][j]=0;
				}
				return nb;
			}
		}
	}

	return nb;
}


int GrilleSudoku::nbSolutions2(int nb_a_resoudre, int nb_resolu, int i, int j)
{
	int k,nb=0,n;
	int liste[9];
	if (nb_a_resoudre>nb_resolu)
	{
		while(i<9)
		{
			while(j<9)
			{
				if (nombresPossibles(liste, i, j, &n)) //si m[i][j]==0
				{
					for (k=0;k<n;k++)
					{
						m[i][j]=liste[k];
						nb += nbSolutions2(nb_a_resoudre,nb_resolu+1,i,j);
						m[i][j]=0;
					}
					return nb;
				}
				j++;
			}
			i++;
		}
	}
	else nb++;
	return nb;
}

/*
int nbCasesVides()
Retourne le nombre de cases vides (m[i][j]==0)
*/
int GrilleSudoku::nbCasesVides()
{
	int i,j,nb=0;
	for (i=0;i<9;i++)
		for (j=0;j<9;j++)
			if(m[i][j]==0) nb++;
	return nb;
}


/*
void ligne(int* liste, int i)
Remplit liste avec les nombres de la ligne de coordonnes(i)
	int* liste	: liste qui va contenir les nombres
	int i		: indice ligne
*/
void GrilleSudoku::ligne(int* liste, int i){for (int j=0;j<9;j++) liste[j]=m[i][j];}

/*
void colonne(int* liste, int j)
Remplit liste avec les nombres de la colonne de coordonne (j)
	int* liste	: liste qui va contenir les nombres
	int j		: indice colonne
*/
void GrilleSudoku::colonne(int* liste, int j){for (int i=0;i<9;i++) liste[i]=m[i][j];}

/*
void zone(int* liste, int i,int j)
Remplit liste avec les nombres de la zone ou se trouve la case de coordonnes (i;j)
	int* liste	: liste qui va contenir les nombres
	int i		: indice ligne
	int j		: indice colonne
*/
void GrilleSudoku::zone(int* liste, int i,int j)
{
	i=i/3*3;
	j=j/3*3;
	for (int k=0;k<9;k++) liste[k]=m[i+k%3][j+k/3];
}

/*
bool nEstPasDans(int* liste,int n)
Retourne true si n ne se trouve pas dans liste sinon false
	int* liste	: liste de nombres
	int n		: nombre  chercher
*/
bool GrilleSudoku::nEstPasDans(int* liste,int n)
{
	int k;
	for (k=0;k<9;k++)
	{
		if(liste[k]==n) return false;
	}
	return true;
}

/*
bool nombresPossibles(int* liste, int i, int j, int* n)
Remplit liste avec les nombres possibles d'une case.
Retourne si la case est vide ou non : 	false -> vide
										true -> non vide
	int* liste	: liste qui va contenir les nombres possibles
	int i		: indice ligne
	int j		: indice colonne
	int* n		: contiendra le nombre de nombres possibles
*/
bool GrilleSudoku::nombresPossibles(int* liste, int i, int j, int* n)
{
	int l1[9],l2[9],l3[9];
	int k,indice=0;
	if(m[i][j]!=0) return false;
	for(k=0;k<9;k++) l1[k]=l2[k]=l3[k]=liste[k]=0;
	ligne(l1,i); //rcupre la ligne
	colonne(l2,j); //rcupre la colonne
	zone(l3,i,j); //rcupre la zone
	for(k=1;k<10;k++)
	{
		if (nEstPasDans(l1,k) && nEstPasDans(l2,k) && nEstPasDans(l3,k))
		{
			liste[indice++]=k;
		}
	}
	*n=indice;
	return true;
}

/*
void setMLine(int nb,int i,int j)
Fixe la valeur de mLine  nb en (i;j) et met  jour le status (couleurs)
	int nb	: valeur
	int i	: indice ligne
	int j	: indice colonne
*/
void GrilleSudoku::setMLine(int nb,int i,int j)
{
     QString str;

     if (nb==0 || status[i][j]==0)
     {
        if (nb==0) mLine[i][j]->setText(QString::null);
        else mLine[i][j]->setText(str.setNum(nb));
        mLine[i][j]->setReadOnly ( false );

        QPalette newPalette = palette();
        newPalette.setColor(QPalette::Base, *normalColor);
        mLine[i][j]->setPalette(newPalette);
     }
     else
     {
        mLine[i][j]->setText(str.setNum(nb));
        mLine[i][j]->setReadOnly ( true );

        QPalette newPalette = palette();

        if (status[i][j]==1)
        {
			newPalette.setColor(QPalette::Base, *genererColor);
		}
		else if (status[i][j]==2)
		{
			newPalette.setColor(QPalette::Base, *resoluColor);
		}
		else if (status[i][j]==3)
		{
			newPalette.setColor(QPalette::Base, *aideColor);
		}
        mLine[i][j]->setPalette(newPalette);
     }

     setNombresPossibles(nbPossibles);
}




void GrilleSudoku::resoudreRecursivement(int nb_a_resoudre, int nb_resolu)
{
	int i,j,k,n;
	int liste[9];
	bool solution=true;

	if (nb_a_resoudre>nb_resolu)
	{
		for (i=0;i<9 && solution;i++)
		{
			for (j=0;j<9 && solution;j++)
			{
				if (nombresPossibles(liste, i, j, &n)) //si m[i][j]==0
				{
					solution=false;
					for (k=0;k<n;k++)
					{
						m[i][j]=liste[k];
						resoudreRecursivement(nb_a_resoudre,nb_resolu+1);
						m[i][j]=0;
					}
				}
			}
		}
	}
	else
	{
		for (i=0;i<9;i++) for (j=0;j<9;j++) m_resultat[i][j]=m[i][j];
		jeu_resolu=true;
	}
}

bool GrilleSudoku::testNbSolutions()
{
	bool d;
	delaiOK = true;
	delai->start(10000);
	if (nbCasesVides()>65 || nombreSolutions()>1)
	{
		delai->stop();
		if (delaiOK) QMessageBox::warning(this, tr("Erreur"),tr("Cette grille comporte plusieurs solutions !"));
		return false;
	}
	delai->stop();
	d=delaiOK;
	delaiOK = true;
	return d;
}

/*************************************/
/************public slots*************/
/*************************************/
void GrilleSudoku::setNbPossibles()
{
	setNombresPossibles(nbPossibles);
}

void GrilleSudoku::delaiPasse()
{
	int i,j;
	QMessageBox::warning(this, tr("Erreur"),tr("Cette opration a t stoppe, il est probable que cette grille ne soit pas correcte."));

	for (i=0;i<9;i++) for (j=0;j<9;j++) m[i][j] = m_debut[i][j];

	delaiOK = false;
}
